/* Pavel Surynek, 2008 */
/* bibox.h */

/*
   This is an implementation of the multi-robot path planning algorithm for biconnected environments with
   at least two free places. The code is not commented however it is simple enough to be understood
   without comments.

   The problem is given as an undirected graph, where each vertex of the graph is either free or
   occupied by a robot. A robot can move to a neighboring unoccupied vertex supposed that no other
   robot is simultaneously entering the target vertex. The task is to move robots preserving this
   condition to their final positions.

   This program is an implementation of the algorithm for solving the problem where the graph is
   biconnected and threre are at least two unoccupied vertices in the graph. The program runs in
   polynomial time with respect to the size of the input graph. More precisely it runs in O(|V|^3)
   where G=(V,E) is the input biconnected graph.
 */

#ifndef __BIBOX_H__
#define __BIBOX_H__

#define UNUSED(X)


#include <vector>
#include <list>
#include <set>
#include <map>


#include "version.h"

class FINE_MOVE
{
public:
	FINE_MOVE()
	{
		m_x = m_y = 0;
		m_time = 0;
	}

	FINE_MOVE(int x, int y)
	{
		m_x = x;
		m_y = y;
		m_time = 0;
	}

	FINE_MOVE(int x, int y, int  time)
	{
		m_x = x;
		m_y = y;
		m_time = time;
	}

	FINE_MOVE(const FINE_MOVE &fm)
	{
		m_x = fm.m_x;
		m_y = fm.m_y;
		m_time = fm.m_time;
	}

	const FINE_MOVE& operator=(const FINE_MOVE &fm)
	{
		m_x = fm.m_x;
		m_y = fm.m_y;
		m_time = fm.m_time;
		
		return *this;
	}

	void reverse(void)
	{
		int t = m_x;
		m_x = m_y;
		m_y = t;
	}

	void save(FILE *f) const
	{
		fprintf(f, "%d ---> %d (%d)\n", m_x, m_y, m_time);
	}

	void load(FILE *f)
	{
		fscanf(f, "%d ---> %d (%d)\n", &m_x, &m_y, &m_time);
	}

	static inline bool dependent(const FINE_MOVE &m1, const FINE_MOVE &m2)
	{
		if (m1.m_x == m2.m_x || m1.m_x == m2.m_y || m1.m_y == m2.m_x || m1.m_y == m2.m_y)
		{
			return true;
		}
		return false;
	} 

public:
	int m_x, m_y;
	int m_time;
};


typedef std::list<FINE_MOVE> FINE_SOLUTION;


typedef std::list<int> IDS_list;
typedef std::set<int, std::less<int> > IDS_set;

extern int MOVE_COUNT;
extern int STEP_COUNT;
extern int DATABASE_HIT_COUNT;
extern int DATABASE_TOTAL_COUNT;
extern int COMP_SOLVED_COUNT;
extern int COMP_TOTAL_COUNT;

extern bool INDEPENDENT;
extern int CURRENT_CYCLE;

class CYCLE
{
public:
	int m_x_id, m_y_id;
	IDS_list m_vertices;
	IDS_list m_loop;
	IDS_list m_bridge;
};


typedef std::vector<CYCLE> CYCLES_vector;


class VERTEX
{
public:
	bool operator==(const VERTEX &vertex) const
	{
		return (m_id == vertex.m_id);
	}

	bool operator<(const VERTEX &vertex) const
	{
		return (m_id < vertex.m_id);
	}

	void print(void) const
	{
		printf("(%d:%d)[%d:%d:%d] ", m_id, m_cyc_id, m_robot, m_final, m_goal);

		for (IDS_list::const_iterator neighbor = m_neighbors.begin(); neighbor != m_neighbors.end(); ++neighbor)
		{
			printf("%d ", *neighbor);
		}
		printf("\n");
	}
public:
	int m_id;
	IDS_list m_neighbors;

// Shortest path specific members
	bool m_active;
	int m_val;
	int m_prev_id;

// Cycle decomposition specific members
	int m_cyc_id;

// Robots specific members
	int m_robot;
	int m_final;
	int m_goal;

// Algorithm specific members
	int m_affected;
};


class EDGE
{
public:
	EDGE()
	{
		m_U_id = 0;
		m_V_id = 0;
		m_cyc_id = 0;
	}

	EDGE(int U_id, int V_id)
	{
		m_U_id = U_id;
		m_V_id = V_id;
		m_cyc_id = 0;
	}

	EDGE(const EDGE& edge)
	{
		m_U_id = edge.m_U_id;
		m_V_id = edge.m_V_id;
		m_cyc_id = edge.m_cyc_id;
	}

	const EDGE& operator=(const EDGE& edge)
	{
		m_U_id = edge.m_U_id;
		m_V_id = edge.m_V_id;
		m_cyc_id = edge.m_cyc_id;

		return *this;
	}

	bool operator==(const EDGE &edge) const
	{
		return (m_U_id == edge.m_U_id && m_V_id == edge.m_V_id);
	}

	bool operator<(const EDGE &edge) const
	{
		return ((m_U_id < edge.m_U_id) ? true : ((m_U_id > edge.m_U_id) ? false : ((m_V_id < edge.m_V_id) ? true : ((m_V_id > edge.m_V_id) ? false : false))));
	}

	void print(void) const
	{
		printf("{%d,%d}\n", m_U_id, m_V_id);
	}

public:
	int m_U_id;
	int m_V_id;

// 	Cycle decomposition specific members
	int m_cyc_id;
};


typedef std::map<int, VERTEX, std::less<int> > VERTICES_map;
typedef std::set<EDGE, std::less<EDGE> > EDGES_set;

typedef std::set<int, std::less<int> > VERTICES_set;


enum ORIENTATION
{
	UN,
	CW,
	CCW
};


class GRAPH
{
public:
	void add_vertex(int id)
	{
		VERTEX &vertex = m_vertices[id];

		vertex.m_id = id;
		vertex.m_cyc_id = 0;
		vertex.m_robot = 0;
		vertex.m_final = 0;
		vertex.m_affected = -1;
	}


	void add_vertex(int id, int cyc_id)
	{
		VERTEX &vertex = m_vertices[id];

		vertex.m_id = id;
		vertex.m_cyc_id = cyc_id;
		vertex.m_robot = 0;
		vertex.m_final = 0;
		vertex.m_goal = 0;
		vertex.m_affected = -1;
	}

	void add_vertex(int id, int cyc_id, int robot, int final)
	{
		VERTEX &vertex = m_vertices[id];

		vertex.m_id = id;
		vertex.m_cyc_id = cyc_id;
		vertex.m_robot = robot;
		vertex.m_final = final;
		vertex.m_goal = final;
		vertex.m_affected = -1;
	}

	void add_edge(int U_id, int V_id)
	{
		EDGE e;
		e.m_U_id = U_id;
		e.m_V_id = V_id;
		e.m_cyc_id = 0;

		m_edges.insert(e);
		m_vertices[U_id].m_neighbors.push_back(V_id);
		m_vertices[V_id].m_neighbors.push_back(U_id);
	}

	void add_edge(int U_id, int V_id, int cyc_id)
	{
		EDGE e;
		e.m_U_id = U_id;
		e.m_V_id = V_id;
		e.m_cyc_id = cyc_id;

		m_edges.insert(e);
		m_vertices[U_id].m_neighbors.push_back(V_id);
		m_vertices[V_id].m_neighbors.push_back(U_id);
	}

	void print(void) const
	{
		printf("V = \n");
		for (VERTICES_map::const_iterator vertex = m_vertices.begin(); vertex != m_vertices.end(); ++vertex)
		{
			vertex->second.print();
		}

		printf("E = \n");
		for (EDGES_set::const_iterator edge = m_edges.begin(); edge != m_edges.end(); ++edge)
		{
			edge->print();
		}

		printf("C = \n");
		unsigned int i;
		for (i = 0; i < m_cycles.size(); ++i)
		{
			printf("%d (%d,%d): ", i, m_cycles[i].m_x_id, m_cycles[i].m_y_id);

			for (std::list<int>::const_iterator vertex = m_cycles[i].m_vertices.begin(); vertex != m_cycles[i].m_vertices.end(); ++vertex)
			{
				printf("%d ", *vertex);
			}
			printf(" [");
			for (std::list<int>::const_iterator vertex = m_cycles[i].m_loop.begin(); vertex != m_cycles[i].m_loop.end(); ++vertex)
			{
				printf("%d ", *vertex);
			}
			printf("]");
			printf(" {");
			for (std::list<int>::const_iterator vertex = m_cycles[i].m_bridge.begin(); vertex != m_cycles[i].m_bridge.end(); ++vertex)
			{
				printf("%d ", *vertex);
			}
			printf("}\n");
		}
	}

public:
	VERTICES_map m_vertices;
	EDGES_set m_edges;

// Cycle decomposition specific members
	CYCLES_vector m_cycles;
	FINE_SOLUTION GLOBAL_SOLUTION;
};


extern std::list<std::pair<int, int> > Independent;

void swap_robots(int a_id, int b_id, GRAPH *graph);
bool swap_robots_empty(int a_id, int b_id, GRAPH *graph);

void swap_finals(int a_id, int b_id, GRAPH *graph);
void swap_goals(int a_id, int b_id, GRAPH *graph);
VERTICES_map::const_iterator find_robot(int robot, GRAPH *graph);
VERTICES_map::const_iterator find_final(int final, GRAPH *graph);
void rotate_cycle_CW(int free, CYCLE *cycle, GRAPH *graph);
void rotate_cycle_CW(CYCLE *cycle, GRAPH *graph);
void rotate_cycle_CCW(int free, CYCLE *cycle, GRAPH *graph);
void rotate_cycle_CCW(CYCLE *cycle, GRAPH *graph);
void half_rotate_cycle_CW(int robot, int id, int free, CYCLE *cycle, GRAPH *graph);
void half_rotate_cycle_CW(int robot, int id, CYCLE *cycle, GRAPH *graph);
void half_rotate_cycle_CCW(int robot, int id, int free, CYCLE *cycle, GRAPH *graph);
void half_rotate_cycle_CCW(int robot, int id, CYCLE *cycle, GRAPH *graph);
ORIENTATION calc_orientation(int U_id, int V_id, CYCLE *cycle);
void find_shortest_path(int source_id, int sink_id, GRAPH *graph, IDS_set &fvertices, IDS_set &fcycles);
int find_nearest_outsider(int source_id, CYCLE *cycle, GRAPH *graph, IDS_set &fvertices, IDS_set &fcycles);
void find_shortest_path(int source_id, int sink_id, GRAPH *graph);
int find_nearest_outsider(int source_id, CYCLE *cycle, GRAPH *graph);
int find_nearest_free(int source_id, GRAPH *graph, IDS_set &fvertices, IDS_set &fcycles);
int find_nearest_free(int source_id, GRAPH *graph, IDS_set &fvertices, IDS_set &fcycles, IDS_list &floop);
void move_free(int init_id, int goal_id, GRAPH *graph, IDS_set &fvertices, IDS_set &fcycles);
void move_free(int init_id, int goal_id, GRAPH *graph);
int find_nearest(CYCLE *cycle, GRAPH *graph);
EDGES_set::iterator find_edge(int U_id, int V_id, GRAPH *graph);
void move_robot(int robot, int where_id, int free, GRAPH *graph, IDS_set &fvertices, IDS_set &fcycles);
void move_robot(int robot, int where_id, GRAPH *graph, IDS_set &fvertices, IDS_set &fcycles);
int find_next_robot_CW(int robot, CYCLE *cycle, GRAPH *graph);
int find_first_final_CW(CYCLE *cycle, GRAPH *graph);
int find_next_final_CW(int robot, CYCLE *cycle, GRAPH *graph);
int find_next_robot_CCW(int robot, CYCLE *cycle, GRAPH *graph);
void exchange_last_cycle(int robot1, int robot2, GRAPH *graph, IDS_set &UNUSED(fvertices), IDS_set &UNUSED(fcycles));
bool solve_critical_last_free(GRAPH *graph, IDS_set &fcycles);
bool solve_critical_free(GRAPH *graph);
bool solve_critical(GRAPH *UNUSED(graph));
void move_robot(int robot, int where, GRAPH *graph);
void generate_bipartite(int c, int m, int l, int f, GRAPH *graph);
void export_graph_2_PDDL(const char *problem, const char *filename, GRAPH *graph);
void generate_bipartite(int m, int l, GRAPH *graph);

#endif
